@@ -1,20 +1,26 @@
 use regex::Regex;
-use std::collections::hash_map::DefaultHasher;
-use std::hash::Hasher;
 use std::io::prelude::*;
-use std::str::FromStr;
 use std::{
-    collections::{HashMap, HashSet},
+    cmp::Reverse,
+    collections::hash_map::DefaultHasher,
+    collections::HashMap,
     env,
-    hash::Hash,
+    hash::{Hash, Hasher},
     path::{Path, PathBuf},
-    sync::{Arc, Mutex},
+    str::FromStr,
+    sync::{
+        mpsc::{
+            self, {Receiver, Sender},
+        },
+        Arc, Mutex,
+    },
     time::Instant,
 };
 
 use chrono::prelude::*;
 use chrono::DateTime;
 use crypto_msg_parser::{extract_symbol, parse_l2, parse_trade, MarketType, MessageType};
+use dashmap::{DashMap, DashSet};
 use flate2::write::GzEncoder;
 use flate2::{read::GzDecoder, Compression};
 use glob::glob;
@@ -22,8 +28,6 @@ use log::*;
 use rlimit::{setrlimit, Resource};
 use serde::{Deserialize, Serialize};
 use serde_json::Value;
-use std::sync::mpsc;
-use std::sync::mpsc::{Receiver, Sender};
 use threadpool::ThreadPool;
 
 #[derive(Serialize, Deserialize)]
@@ -50,14 +54,19 @@ fn get_day(unix_timestamp: i64) -> String {
 #[derive(Clone)]
 #[allow(clippy::type_complexity)]
 struct Output(
-    Arc<
-        Mutex<(
-            Box<dyn std::io::Write + Send>,
-            Box<dyn std::io::Write + Send>,
-        )>,
-    >,
+    Arc<(
+        Mutex<Box<dyn std::io::Write + Send>>,
+        Mutex<Box<dyn std::io::Write + Send>>,
+    )>,
 );
 
+fn is_blocked_market(market_type: MarketType) -> bool {
+    market_type == MarketType::QuantoFuture
+        || market_type == MarketType::QuantoSwap
+        || market_type == MarketType::Move
+        || market_type == MarketType::BVOL
+}
+
 /// Split a file by symbol and write to multiple files.
 ///
 /// This function does split, dedup and parse together, and it is
@@ -78,8 +87,8 @@ fn split_file<P>(
     day: String,
     output_dir_raw: P,
     output_dir_parsed: P,
-    split_files: Arc<Mutex<HashMap<String, Output>>>,
-    visited: Arc<Mutex<HashSet<u64>>>,
+    split_files: Arc<DashMap<String, Output>>,
+    visited: Arc<DashSet<u64>>,
 ) -> (i64, i64)
 where
     P: AsRef<Path>,
@@ -110,14 +119,12 @@ where
                         msg.json.hash(&mut hasher);
                         hasher.finish()
                     };
-                    let mut visited = visited.lock().unwrap();
                     visited.insert(hashcode)
                 };
                 if is_new {
                     if let Some(symbol) = extract_symbol(exchange, market_type, &msg.json) {
                         let output = {
-                            let mut m = split_files.lock().unwrap();
-                            if !m.contains_key(&symbol) {
+                            if !split_files.contains_key(&symbol) {
                                 let buf_writer_raw = {
                                     let output_file_name = format!(
                                         "{}.{}.{}.{}.{}.json.gz",
@@ -141,6 +148,7 @@ where
                                         Compression::default(),
                                     ))
                                 };
+
                                 let buf_writer_parsed = {
                                     let pair =
                                         crypto_pair::normalize_pair(&symbol, exchange).unwrap();
@@ -168,29 +176,25 @@ where
                                         Compression::default(),
                                     ))
                                 };
-                                m.insert(
+                                split_files.insert(
                                     symbol.clone(),
-                                    Output(Arc::new(Mutex::new((
-                                        Box::new(buf_writer_raw),
-                                        Box::new(buf_writer_parsed),
-                                    )))),
+                                    Output(Arc::new((
+                                        Mutex::new(Box::new(buf_writer_raw)),
+                                        Mutex::new(Box::new(buf_writer_parsed)),
+                                    ))),
                                 );
                             }
-                            m.get(&symbol).unwrap().clone()
+                            let entry = split_files.get(&symbol).unwrap();
+                            entry.value().clone()
                         };
-                        let mut writers = output.0.lock().unwrap();
                         // raw
                         if day == get_day((msg.received_at / 1000_u64) as i64) {
-                            writeln!(writers.0, "{}", line).unwrap();
+                            writeln!(output.0 .0.lock().unwrap(), "{}", line).unwrap();
                         }
                         match msg.msg_type {
                             MessageType::L2Event => {
                                 // Skip unsupported markets
-                                if market_type != MarketType::QuantoFuture
-                                    && market_type == MarketType::QuantoSwap
-                                    && market_type == MarketType::Move
-                                    && market_type == MarketType::BVOL
-                                {
+                                if !is_blocked_market(market_type) {
                                     if let Ok(messages) = parse_l2(
                                         exchange,
                                         msg.market_type,
@@ -200,7 +204,7 @@ where
                                         for message in messages {
                                             if get_day(message.timestamp / 1000) == day {
                                                 writeln!(
-                                                    writers.1,
+                                                    output.0 .1.lock().unwrap(),
                                                     "{}",
                                                     serde_json::to_string(&message).unwrap()
                                                 )
@@ -217,7 +221,7 @@ where
                                     for message in messages {
                                         if get_day(message.timestamp / 1000) == day {
                                             writeln!(
-                                                writers.1,
+                                                output.0 .1.lock().unwrap(),
                                                 "{}",
                                                 serde_json::to_string(&message).unwrap()
                                             )
@@ -245,7 +249,7 @@ where
     (error_lines, total_lines)
 }
 
-fn sort_file<P>(input_file: P, output_file: P) -> (i64, i64)
+fn sort_file<P>(input_file: P, output_file: P, use_pixz: bool) -> (i64, i64)
 where
     P: AsRef<Path>,
 {
@@ -284,20 +288,55 @@ where
     if error_lines == 0 {
         lines.sort_by_key(|x| x.0); // sort by timestamp
 
-        let mut writer = {
-            let f_out = std::fs::OpenOptions::new()
-                .create(true)
-                .write(true)
-                .truncate(true)
-                .open(output_file)
-                .unwrap();
-            let e = xz2::write::XzEncoder::new(f_out, 9);
-            std::io::BufWriter::new(e)
-        };
-        for line in lines {
-            writeln!(writer, "{}", line.1).unwrap();
+        if !use_pixz {
+            let mut writer = {
+                let f_out = std::fs::OpenOptions::new()
+                    .create(true)
+                    .write(true)
+                    .truncate(true)
+                    .open(output_file)
+                    .unwrap();
+                let e = xz2::write::XzEncoder::new(f_out, 9);
+                std::io::BufWriter::new(e)
+            };
+            for line in lines {
+                writeln!(writer, "{}", line.1).unwrap();
+            }
+            writer.flush().unwrap();
+        } else {
+            let json_file = {
+                let output_dir = output_file.as_ref().parent().unwrap().to_path_buf();
+                let filename = output_file.as_ref().file_name().unwrap().to_str().unwrap();
+                output_dir.join(&filename[..filename.len() - 3])
+            };
+            let mut writer = {
+                let f_out = std::fs::OpenOptions::new()
+                    .create(true)
+                    .write(true)
+                    .truncate(true)
+                    .open(json_file.as_path())
+                    .unwrap();
+                std::io::BufWriter::new(f_out)
+            };
+            for line in lines {
+                writeln!(writer, "{}", line.1).unwrap();
+            }
+            writer.flush().unwrap();
+            drop(writer);
+            match std::process::Command::new("pixz")
+                .args(["-9", json_file.as_path().to_str().unwrap()])
+                .output()
+            {
+                Ok(output) => {
+                    if output.status.success() {
+                        debug!("{}", String::from_utf8_lossy(&output.stdout));
+                    } else {
+                        panic!("{}", String::from_utf8_lossy(&output.stderr));
+                    }
+                }
+                Err(err) => panic!("{}", err),
+            }
         }
-        writer.flush().unwrap();
         std::fs::remove_file(input_file).unwrap();
     }
     (error_lines, total_lines)
@@ -307,7 +346,9 @@ where
 ///
 /// Each `(exchange, msg_type, market_type, day)` will launch a process.
 fn process_files_of_day(
+    exchange: &str,
     msg_type: MessageType,
+    market_type: MarketType,
     day: &str,
     input_dir: &str,
     output_dir_raw: &str,
@@ -318,7 +359,10 @@ fn process_files_of_day(
 
     // split
     {
-        let glob_pattern = format!("{}/**/*.*.{}.{}-??-??.json.gz", input_dir, msg_type, day);
+        let glob_pattern = format!(
+            "{}/**/{}.{}.{}.{}-??-??.json.gz",
+            input_dir, exchange, market_type, msg_type, day
+        );
         let mut paths: Vec<PathBuf> = glob(&glob_pattern)
             .unwrap()
             .filter_map(Result::ok)
@@ -341,8 +385,8 @@ fn process_files_of_day(
             };
             let mut paths_of_next_day: Vec<PathBuf> = glob(
                 format!(
-                    "{}/**/*.*.{}.{}-??.json.gz",
-                    input_dir, msg_type, next_day_first_hour
+                    "{}/**/{}.{}.{}.{}-??.json.gz",
+                    input_dir, exchange, market_type, msg_type, next_day_first_hour
                 )
                 .as_str(),
             )
@@ -352,21 +396,23 @@ fn process_files_of_day(
             paths.append(&mut paths_of_next_day);
         }
 
-        info!("Started split {} {}", msg_type, day);
+        info!(
+            "Started split {} {} {} {}",
+            exchange, market_type, msg_type, day
+        );
         #[allow(clippy::type_complexity)]
         let (tx, rx): (Sender<(i64, i64)>, Receiver<(i64, i64)>) = mpsc::channel();
         let start_timstamp = Instant::now();
-        // small files get processed first
-        paths.sort_by_cached_key(|path| std::fs::metadata(path).unwrap().len());
-        // key = exchange.market_type.msg_type
-        let mut visited_map: HashMap<String, Arc<Mutex<HashSet<u64>>>> = HashMap::new();
-        let mut split_files_map: HashMap<String, Arc<Mutex<HashMap<String, Output>>>> =
-            HashMap::new();
+        // Larger files get processed first
+        paths.sort_by_cached_key(|path| Reverse(std::fs::metadata(path).unwrap().len()));
+        let visited: Arc<DashSet<u64>> = Arc::new(DashSet::new());
+        let split_files: Arc<DashMap<String, Output>> = Arc::new(DashMap::new());
         for path in paths {
             let file_name = path.as_path().file_name().unwrap();
             let v: Vec<&str> = file_name.to_str().unwrap().split('.').collect();
-            let exchange = v[0];
+            assert_eq!(exchange, v[0]);
             let market_type_str = v[1];
+            assert_eq!(market_type, MarketType::from_str(market_type_str).unwrap());
             let msg_type_str = v[2];
             assert_eq!(msg_type, MessageType::from_str(msg_type_str).unwrap());
             let output_dir_raw = Path::new(output_dir_raw)
@@ -379,16 +425,9 @@ fn process_files_of_day(
                 .join(exchange)
                 .join(market_type_str);
             std::fs::create_dir_all(output_dir_parsed.as_path()).unwrap();
-            let key = format!("{}.{}.{}", exchange, market_type_str, msg_type_str);
-            if !visited_map.contains_key(&key) {
-                visited_map.insert(key.clone(), Arc::new(Mutex::new(HashSet::new())));
-            }
-            if !split_files_map.contains_key(&key) {
-                split_files_map.insert(key.clone(), Arc::new(Mutex::new(HashMap::new())));
-            }
 
-            let visited_of_market = visited_map.get(&key).unwrap().clone();
-            let split_files_of_market = split_files_map.get(&key).unwrap().clone();
+            let visited_clone = visited.clone();
+            let split_files_clone = split_files.clone();
             let day_clone = day.to_string();
             let thread_tx = tx.clone();
             thread_pool.execute(move || {
@@ -397,8 +436,8 @@ fn process_files_of_day(
                     day_clone,
                     output_dir_raw.as_path(),
                     output_dir_parsed.as_path(),
-                    split_files_of_market,
-                    visited_of_market,
+                    split_files_clone,
+                    visited_clone,
                 );
                 thread_tx.send(t).unwrap();
             });
@@ -411,18 +450,18 @@ fn process_files_of_day(
             error_lines += t.0;
             total_lines += t.1;
         }
-        for m in split_files_map.values() {
-            for output in m.lock().unwrap().values() {
-                let mut output = output.0.lock().unwrap();
-                output.0.flush().unwrap();
-                output.1.flush().unwrap();
-            }
+        for entry in split_files.iter() {
+            let output = entry.value();
+            output.0 .0.lock().unwrap().flush().unwrap();
+            output.0 .1.lock().unwrap().flush().unwrap();
         }
         let error_ratio = (error_lines as f64) / (total_lines as f64);
         if error_ratio > 0.01 {
             // error ratio > 1%
             error!(
-                "Failed to split {} {}, because error ratio {}/{}={}% is higher than 1% !",
+                "Failed to split {} {} {} {}, because error ratio {}/{}={}% is higher than 1% !",
+                exchange,
+                market_type,
                 msg_type,
                 day,
                 error_lines,
@@ -431,13 +470,16 @@ fn process_files_of_day(
             );
             return false;
         } else {
-            info!("Finished split {} {}, dropped {} malformed lines out of {} lines, time elapsed {} seconds", msg_type, day, error_lines, total_lines, start_timstamp.elapsed().as_secs());
+            info!("Finished split {} {} {} {}, dropped {} malformed lines out of {} lines, time elapsed {} seconds", exchange, market_type, msg_type, day, error_lines, total_lines, start_timstamp.elapsed().as_secs());
         }
     }
 
     // sort by timestamp
     {
-        let glob_pattern = format!("{}/**/*.*.{}.*.{}.json.gz", output_dir_raw, msg_type, day);
+        let glob_pattern = format!(
+            "{}/**/{}.{}.{}.*.{}.json.gz",
+            output_dir_raw, exchange, market_type, msg_type, day
+        );
         let paths_raw: Vec<PathBuf> = glob(&glob_pattern)
             .unwrap()
             .filter_map(Result::ok)
@@ -448,29 +490,38 @@ fn process_files_of_day(
         }
 
         let glob_pattern = format!(
-            "{}/**/*.*.{}.*.{}.json.gz",
-            output_dir_parsed, msg_type, day
+            "{}/**/{}.{}.{}.*.{}.json.gz",
+            output_dir_parsed, exchange, market_type, msg_type, day
         );
         let paths_parsed: Vec<PathBuf> = glob(&glob_pattern)
             .unwrap()
             .filter_map(Result::ok)
             .collect();
-        if paths_parsed.is_empty() {
-            warn!("There are no files of pattern {}", glob_pattern);
-            return false;
-        }
 
-        info!("Started sort {} {}", msg_type, day);
+        let mut paths = if is_blocked_market(market_type) {
+            for path in paths_parsed {
+                std::fs::remove_file(path).unwrap();
+            }
+            paths_raw
+        } else {
+            paths_raw
+                .into_iter()
+                .chain(paths_parsed.into_iter())
+                .collect()
+        };
+
+        info!(
+            "Started sort {} {} {} {}",
+            exchange, market_type, msg_type, day
+        );
         #[allow(clippy::type_complexity)]
         let (tx, rx): (Sender<(i64, i64)>, Receiver<(i64, i64)>) = mpsc::channel();
-        let mut paths: Vec<PathBuf> = paths_raw
-            .into_iter()
-            .chain(paths_parsed.into_iter())
-            .collect();
         let start_timstamp = Instant::now();
-        // small files get processed first
+        // Smaller files get processed first
         paths.sort_by_cached_key(|path| std::fs::metadata(path).unwrap().len());
-        for input_file in paths {
+        let percentile_90 = ((paths.len() as f64) * 0.9) as usize;
+        let pixz_exists = Path::new("/usr/bin/pixz").exists();
+        for (index, input_file) in paths.into_iter().enumerate() {
             let file_name = input_file.as_path().file_name().unwrap();
             let output_file_name = format!(
                 "{}.json.xz",
@@ -482,10 +533,17 @@ fn process_files_of_day(
             );
             let output_file = Path::new(input_file.parent().unwrap()).join(output_file_name);
             let thread_tx = tx.clone();
-            thread_pool.execute(move || {
-                let t = sort_file(input_file, output_file);
-                thread_tx.send(t).unwrap();
-            });
+            if pixz_exists && index >= percentile_90 {
+                thread_pool.execute(move || {
+                    let t = sort_file(input_file, output_file, true);
+                    thread_tx.send(t).unwrap();
+                });
+            } else {
+                thread_pool.execute(move || {
+                    let t = sort_file(input_file, output_file, false);
+                    thread_tx.send(t).unwrap();
+                });
+            }
         }
         thread_pool.join();
         drop(tx); // drop the sender
@@ -497,7 +555,9 @@ fn process_files_of_day(
         }
         if error_lines == 0 {
             info!(
-                "Finished sort {} {}, time elapsed {} seconds",
+                "Finished sort {} {} {} {}, time elapsed {} seconds",
+                exchange,
+                market_type,
                 msg_type,
                 day,
                 start_timstamp.elapsed().as_secs()
@@ -505,7 +565,9 @@ fn process_files_of_day(
             true
         } else {
             error!(
-                "Failed to sort {} {}, found {} malformed lines out of {} lines, time elapsed {} seconds",
+                "Failed to sort {} {} {} {}, found {} malformed lines out of {} lines, time elapsed {} seconds",
+                exchange,
+                market_type,
                 msg_type,
                 day,
                 error_lines, total_lines,
@@ -521,36 +583,53 @@ fn main() {
     assert!(setrlimit(Resource::NOFILE, 4096, 4096).is_ok());
 
     let args: Vec<String> = env::args().collect();
-    if args.len() != 6 {
-        eprintln!("Usage: crypto-daily-processor <msg_type> <day> <input_dir> <output_dir_raw> <output_dir_parsed>");
+    if args.len() != 8 {
+        eprintln!("Usage: crypto-daily-processor <exchange> <msg_type> <market_type> <day> <input_dir> <output_dir_raw> <output_dir_parsed>");
         std::process::exit(1);
     }
 
-    let msg_type = MessageType::from_str(&args[1]);
+    let exchange: &'static str = Box::leak(args[1].clone().into_boxed_str());
+
+    let msg_type = MessageType::from_str(&args[2]);
     if msg_type.is_err() {
-        eprintln!("Unknown msg type: {}", &args[1]);
+        eprintln!("Unknown msg type: {}", &args[2]);
         std::process::exit(1);
     }
     let msg_type = msg_type.unwrap();
 
-    let day: &'static str = Box::leak(args[2].clone().into_boxed_str());
+    let market_type = MarketType::from_str(&args[3]);
+    if market_type.is_err() {
+        eprintln!("Unknown market type: {}", &args[3]);
+        std::process::exit(1);
+    }
+    let market_type = market_type.unwrap();
+
+    let day: &'static str = Box::leak(args[4].clone().into_boxed_str());
     let re = Regex::new(r"^\d{4}-\d{2}-\d{2}$").unwrap();
     if !re.is_match(day) {
         eprintln!("{} is invalid, day should be yyyy-MM-dd", day);
         std::process::exit(1);
     }
 
-    let input_dir: &'static str = Box::leak(args[3].clone().into_boxed_str());
+    let input_dir: &'static str = Box::leak(args[5].clone().into_boxed_str());
     if !Path::new(input_dir).is_dir() {
         eprintln!("{} does NOT exist", input_dir);
         std::process::exit(1);
     }
-    let output_dir_raw: &'static str = Box::leak(args[4].clone().into_boxed_str());
-    let output_dir_parsed: &'static str = Box::leak(args[5].clone().into_boxed_str());
+    let output_dir_raw: &'static str = Box::leak(args[6].clone().into_boxed_str());
+    let output_dir_parsed: &'static str = Box::leak(args[7].clone().into_boxed_str());
     std::fs::create_dir_all(Path::new(output_dir_raw)).unwrap();
     std::fs::create_dir_all(Path::new(output_dir_parsed)).unwrap();
 
-    if !process_files_of_day(msg_type, day, input_dir, output_dir_raw, output_dir_parsed) {
+    if !process_files_of_day(
+        exchange,
+        msg_type,
+        market_type,
+        day,
+        input_dir,
+        output_dir_raw,
+        output_dir_parsed,
+    ) {
         std::process::exit(1);
     }
 }
